# Copyright 2022 Cognite AS
import pandas as pd

from statsmodels.tsa.arima.model import ARIMA

from ..type_check import check_types
from ..validations import validate_series_is_not_empty


@check_types
def arma(data: pd.Series, ar_order: int = 2, ma_order: int = 2) -> pd.Series:
    """Autoregressive moving average.

    The autoregressive moving average (ARMA) is a popular model used in forecasting. It uses an autoregression (AR)
    analysis to characterize the effect of past values on current values and a moving average to quantify the effect of the
    previous day's error (variation).


    Args:
        data: Time series.
        ar_order: AR order.
            Number of past data points to include in the AR model. Defaults to 2.
        ma_order: MA order.
            Number of terms in the MA model.  Defaults to 2.

    Returns:
        pandas.Series: Smoothed data.
    """
    # TODO: Refactor to accepet time window

    validate_series_is_not_empty(data)

    # Train model on entire dataset and return fit on dataset
    model = ARIMA(data, order=(ar_order, 0, ma_order))
    fit_model = model.fit()
    y_pred = fit_model.predict()

    return y_pred



# Copyright 2022 Cognite AS
import numpy as np
import pandas as pd

from indsl.exceptions import UserValueError
from indsl.type_check import check_types
from indsl.validations import validate_series_has_minimum_length, validate_series_has_time_index


MIN_DATA_PT = 10


@check_types
def find_period(x: pd.Series, method_period: str = "median") -> float:
    """Estimate Data Period.

    Estimates the period of a time series

    Args:
         x: Time series
         method_period: Method to estimate period
            can be 'median' or 'min' of the difference of timestamps
    Returns:
         float: period in nanoseconds

    Raises:
         TypeError: x is not a time series
         TypeError: index of x is not datetime
         ValueError: method_period is not median or min
    """
    validate_series_has_time_index(x)
    if method_period == "median":
        period = np.median(np.diff(x.index.to_numpy(np.int64))).astype(float)
    elif method_period == "min":
        period = np.min(np.diff(x.index.to_numpy(np.int64))).astype(float)
    else:
        raise UserValueError(f"Period calculation method can only be strings: median or min, not {method_period}")
    return period


@check_types
def calculate_completeness(x: pd.Series, time_start: pd.Timestamp, time_end: pd.Timestamp, period: float) -> float:
    """Completeness score.

    Calculates the completeness score of an input time series for a certain date range

    Args:
         x: Time series
         time_start: start time of date range
         time_end: end time of date range
         period: timeseries period in ns

    Returns:
         float: completeness score

    Raises:
         ValueError: completeness score more than 1
    """
    x = x.loc[(x.index >= time_start) & (x.index <= time_end)]

    total_time_span = time_end - time_start

    expected_num_points = int(total_time_span.total_seconds() * 10**9 / period)
    if x.index[0] == time_start:
        expected_num_points = expected_num_points + 1

    completeness = x.count() / expected_num_points
    if completeness > 1:
        raise ValueError(f"Completeness score {completeness}. Change period calculation method.")
    return completeness


@check_types
def completeness_score(
    x: pd.Series, cutoff_good: float = 0.80, cutoff_med: float = 0.60, method_period: str = "median"
) -> str:
    """Completeness score.

    This function determines the completeness of a time series from a completeness score.
    The score is a function of the inferred data sampling period
    (median or minimum of timestamp differences) and the expected total number of data points for the period
    from the sampling frequency.
    The completeness score is defined as the ratio between the actual number of data points
    to the expected number of data points based on the sampling frequency. The completeness
    is categorized as good if the score is above the
    specified cutoff ratio for good completeness. It is medium if the score falls between
    the cutoff ratio for good and medium completeness. It is characterized as poor if the
    score is below the medium completeness ratio.

    Args:
         x: Time series
         cutoff_good: Good cutoff
             Value between 0 and 1. A completeness score above this cutoff value
             indicates good data completeness. Defaults to 0.80.
         cutoff_med: Medium cutoff
             Value between 0 and 1 and lower than the good data completeness
             cutoff. A completeness score above this cutoff and below the good
             completeness cutoff indicates medium data completeness. Data
             with a completeness score below it are categorised as poor data
             completeness. Defaults to 0.60.
         method_period: Method
             Name of the method used to estimate the period of the time series, can be 'median' or 'min'.
             Default is 'median'

    Returns:
         string: Data quality
             The data quality is defined as
             Good, when completeness score >= cutoff_good,
             Medium, when cutoff_med <= completeness score < cutoff_good,
             Poor, when completeness score < cutoff_med

    Raises:
         TypeError: cutoff_good and cutoff_med are not a number
         ValueError: x has less than ten data points
         TypeError: x is not a time series
         TypeError: index of x is not datetime
         ValueError: method_period is not median or min
         ValueError: completeness score more than 1
    """
    validate_series_has_minimum_length(x, MIN_DATA_PT)

    if cutoff_med is not None and cutoff_good is not None and cutoff_med >= cutoff_good:
        raise UserValueError("cutoff_good should be higher than cutoff_med.")

    period = find_period(x, method_period)

    completeness = calculate_completeness(x, x.index[0], x.index[-1], period)

    if completeness >= cutoff_good:
        return f"Good: completeness score is {completeness:.2f}"
    elif completeness >= cutoff_med:
        return f"Medium: completeness score is {completeness:.2f}"
    else:
        return f"Poor: completeness score is {completeness:.2f}"



# Copyright 2021 Cognite AS
from .utils import no_op  # isort:skip
from . import deprecated_functions
from .deprecated_functions import *  # noqa: F403
from .versioning_test import versioning_test_op


TOOLBOX_NAME = "Not listed operations"

__all__ = ["no_op", "versioning_test_op"] + deprecated_functions.__all__



# Copyright 2021 Cognite AS
import pandas as pd

from indsl import versioning
from indsl.ts_utils.ts_utils import time_parse
from indsl.type_check import check_types


@versioning.register(version="1.0", deprecated=True)
@check_types
def sma(data: pd.Series, time_window: str = "60min", min_periods: int = 1) -> pd.Series:
    """Simple moving average (SMA).

    Plain simple average that computes the sum of the values of the observations in a time_window divided by the number of observations in the time_window.
    SMA time series are much less noisy than the original time series. However, SMA time series lag the original time series, which means that changes in the trend are only seen with a delay (lag) of time_window/2.

    Args:
        data: Time series.
        time_window: Window.
            Length of the time period to compute the average. Defaults to '60min'.
            Accepted string format: '3w', '10d', '5h', '30min', '10s'.
            If the user gives a number without unit (such as '60'), it will be considered as the number of minutes.
        min_periods: Minimum samples.
            Minimum number of observations in window required to have a value (otherwise result is NA). Defaults  to 1.

    Returns:
        pandas.Series: Smoothed time series
    """
    time_window = time_parse(time_window, sma.__name__)

    return data.rolling(time_window, min_periods=min_periods).mean()



# Copyright 2021 Cognite AS
import numpy as np
import pandas as pd

from ..ts_utils.utility_functions import (
    generate_step_series,
    modified_z_scores_test,
    normality_assumption_test,
    z_scores_test,
)
from ..type_check import check_types
from ..validations import validate_series_has_time_index, validate_series_is_not_empty


@check_types
def gaps_identification_z_scores(
    data: pd.Series, cutoff: float = 3.0, test_normality_assumption: bool = False
) -> pd.Series:
    """Gaps detection, Z-scores.

    This function detects gaps in the time stamps using `Z-scores <https://en.wikipedia.org/wiki/Standard_score>`_. Z-score stands for
    the number of standard deviations by which the value of a raw score (i.e., an observed value or data point) is
    above or below the mean value of what is being observed or measured. This method assumes that the time step sizes
    are normally distributed. Gaps are defined as time periods where the Z-score is larger than cutoff.

    Args:
        data: Time series
        cutoff: Cut-off
            Time periods are considered gaps if the Z-score is over this cut-off value. Default 3.0.
        test_normality_assumption: Test for normality.
            Raise a warning if the data is not normally distributed.
            The Shapiro-Wilk test is used. The test is only performed if the time series contains less than 5000 data points.
            Default to False.

    Returns:
        pd.Series: Time series
            The returned time series is an indicator function that is 1 where there is a gap, and 0 otherwise.

    Raises:
        UserTypeError: data is not a time series
        UserTypeError: cutoff is not a number
        UserValueError: data is empty
        UserValueError: time series is not normally distributed
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)

    if len(data) < 2:
        return pd.Series([0] * len(data), index=data.index)

    timestamps = data.index.to_numpy(np.int64)
    diff = np.diff(timestamps)

    if test_normality_assumption:
        normality_assumption_test(series=diff, max_data_points=5000, min_p_value=0.05, min_W=0.5)

    is_gap = z_scores_test(diff, cutoff=cutoff, direction="greater")

    # duplicate the last flag value since the flag series should have the same length as the original data points
    flag_series = pd.concat([pd.Series(is_gap, index=data.index[:-1]), pd.Series(is_gap[-1], index=[data.index[-1]])])

    return generate_step_series(flag_series)


@check_types
def gaps_identification_modified_z_scores(data: pd.Series, cutoff: float = 3.5) -> pd.Series:
    """Gaps detection, mod. Z-scores.

    Detect gaps in the time stamps using modified Z-scores. Gaps are defined as time periods
    where the Z-score is larger than cutoff.

    Args:
        data: Time series
        cutoff: Cut-off
            Time-periods are considered gaps if the modified Z-score is over this cut-off value. Default 3.5.

    Returns:
        pd.Series: Time series
            The returned time series is an indicator function that is 1 where there is a gap, and 0 otherwise.

    Raises:
        UserTypeError: data is not a time series
        UserTypeError: cutoff has to be of type float
        UserValueError: data is empty
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)

    if len(data) < 2:
        return pd.Series([0] * len(data), index=data.index)

    timestamps = data.index.to_numpy(np.int64)
    diff = np.diff(timestamps)

    is_gap = modified_z_scores_test(diff, cutoff=cutoff, direction="greater")

    # duplicate the last flag value since the flag series should have the same length as the original data points
    flag_series = pd.concat([pd.Series(is_gap, index=data.index[:-1]), pd.Series(is_gap[-1], index=[data.index[-1]])])

    return generate_step_series(flag_series)


@check_types
def gaps_identification_iqr(data: pd.Series) -> pd.Series:
    """Gaps detection, IQR.

    Detect gaps in the time stamps using the `interquartile range (IQR)
    <https://en.wikipedia.org/wiki/Interquartile_range>`_ method. The IQR is a measure of statistical
    dispersion, which is the spread of the data. Any time steps more than 1.5 IQR above Q3 are considered
    gaps in the data.

    Args:
        data: time series

    Returns:
        pd.Series: time series
            The returned time series is an indicator function that is 1 where there is a gap, and 0 otherwise.

    Raises:
        UserTypeError: data is not a time series
        UserValueError: data is empty
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)

    if len(data) < 2:
        return pd.Series([0] * len(data), index=data.index)

    timestamps = data.index.to_numpy(np.int64)
    diff = np.diff(timestamps)

    percentile25 = np.quantile(diff, 0.25)
    percentile75 = np.quantile(diff, 0.75)

    iqr = percentile75 - percentile25
    upper_limit = percentile75 + 1.5 * iqr
    is_gap = np.where(diff > upper_limit, 1, 0)

    # duplicate the last flag value since the flag series should have the same length as the original data points
    flag_series = pd.concat([pd.Series(is_gap, index=data.index[:-1]), pd.Series(is_gap[-1], index=[data.index[-1]])])

    return generate_step_series(flag_series)


@check_types
def gaps_identification_threshold(data: pd.Series, time_delta: pd.Timedelta = pd.Timedelta("5m")) -> pd.Series:
    """Gaps detection, threshold.

    Detect gaps in the time stamps using a timedelta threshold.

    Args:
        data: time series
        time_delta: Time threshold
            Maximum time delta between points. Defaults to 5min.

    Returns:
        pd.Series: time series
            The returned time series is an indicator function that is 1 where there is a gap, and 0 otherwise.

    Raises:
        UserTypeError: data is not a time series
        UserValueError: data is empty
        UserTypeError: time_delta is not a pd.Timedelta
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)

    if len(data) < 2:
        return pd.Series([0] * len(data), index=data.index)

    last_index = data.index[-1]

    diff = data.index.to_series().diff().shift(-1).dropna()
    is_gap_series = diff > time_delta
    is_gap_series = is_gap_series.astype(int)

    # duplicate the last flag value since the flag series should have the same length as the original data points
    flag_series = pd.concat([is_gap_series, pd.Series(is_gap_series[-1], index=[last_index])])

    return generate_step_series(flag_series)



# Copyright 2022 Cognite AS
import typing

import numpy as np
import pandas as pd

from statsmodels.tsa.holtwinters import ExponentialSmoothing

from indsl import versioning

from ..exceptions import UserRuntimeError, UserValueError
from ..type_check import check_types
from ..validations import validate_series_has_time_index


@versioning.register(version="1.0", deprecated=True)
@check_types
def holt_winters_predictor(
    data: pd.Series,
    seasonal_periods: int,
    seasonality: typing.Literal["add", "mul"] = "add",
    trend: typing.Literal["add", "mul"] = "add",
    dampen_trend: bool = False,
    steps: int = 1,
    train_fraction: float = 0.8,
) -> pd.Series:
    """Triple exponential smoothing.

    This technique (also known as Holt-Winters) can forecast time series data with a trend and
    seasonal variability.
    It works by utilizing exponential smoothing thrice - for the average value, the trend, and the seasonality.
    Values are predicted by combining the effects of these influences.

    Args:
        data: Time series.
        seasonal_periods: Number of periods per cycle.
            The value for the seasonal periods is chosen to denote the number of timesteps within a period,
            e.g., 7*24 for hourly data with a weekly seasonality or 365 for daily data with a yearly pattern. Note!
            A time series that shows a spike every day, but not on Sunday, does not have a daily seasonality, but a
            weekly seasonality!
        seasonality: Seasonality.
            Additive seasonality: Amount of adjustment is constant.
            Multiplicative seasonality: Amount of adjustment varies with the level of the series.
        trend: Trend.
            Additive seasonality: Amount of adjustment is constant.
            Multiplicative seasonality: Amount of adjustment varies with the level of the series.
        dampen_trend: Dampen the trend component.
            If the trend component shall be dampened. This method is useful to predict very far in the future,
            and it is reasonable to assume that the trend will not stay constant but flatten out.
        steps: Steps.
            Number of steps to forecast ahead of the training dataset.
        train_fraction: Fraction.
            Fraction of the input data used for training the model.

    Returns:
        pandas.Series: Prediction.
        Predicted data for the test fraction of the input data (e.g., 1 - train_fraction).

    Raises:
        UserRuntimeError: If an empty time series is passed into the function.
        UserTypeError: If a time series with the wrong index is provided.
        UserValueError: Incorrect values passed into keyword arguments.
    """
    if len(data) < 1:
        raise UserRuntimeError("No data passed to algorithm.")

    validate_series_has_time_index(data)

    if seasonal_periods <= 1:
        raise UserValueError(f"seasonal_periods must be an integer greater than 1. Got {seasonal_periods}")

    if not 0 < train_fraction < 1:
        raise UserValueError(f"train_fraction needs to be a float between 0 and 1. Got {train_fraction}")

    if trend == "mul" or seasonality == "mul":
        if any(data <= 0):
            raise UserValueError("""When using "mul" for trend or seasonality components, data must be strictly > 0""")

    n_obs = len(data)
    n_train = int(n_obs * train_fraction)
    dt_step_sec = int(data.reset_index()["index"].diff().mean().total_seconds())

    train_data = data.iloc[:n_train]
    pred_data = _train_and_return_forecast(
        train_data, seasonality, trend, seasonal_periods, dampen_trend, steps, dt_step_sec
    )

    return pred_data.dropna()


@check_types
def _train_and_return_forecast(
    train_data: pd.Series,
    seasonality: str,
    trend: str,
    seasonal_periods: int,
    dampen_trend: bool,
    steps: int,
    dt_step_sec: int,
) -> pd.Series:
    """Train and return forecast.

    Function to train an Holt-Winters model and return the forecasted value
    as a Pandas series with a datetime index.

    Args:
        train_data: data to train model on
        seasonality: Either "mul" or "add" - additive or multiplicative seasonality
        trend: Either "mul" or "add" - additive or multiplicative trend
        seasonal_periods: Number of periods in a complete seasonal cycle
        dampen_trend: If the trend should be dampened
        steps: number of steps to forecast ahead of training data
        dt_step_sec: expected granularity for each step (seconds)

    Returns:
        pandas.Series: Forecasted time series of length `steps`
    """
    fitted_model = ExponentialSmoothing(
        train_data,
        seasonal=seasonality,
        trend=trend,
        seasonal_periods=seasonal_periods,
        damped_trend=dampen_trend,
        initialization_method="heuristic",
    ).fit()
    y_pred = fitted_model.forecast(steps).values

    max_dt = train_data.index[-1]
    dt_index = max_dt + pd.TimedeltaIndex(np.linspace(1, steps, num=steps) * dt_step_sec, unit="s")

    return pd.Series(y_pred, index=dt_index)



# Copyright 2022 Cognite AS
import operator as op

from typing import List, Union

import numpy as np

# Simple operations
import pandas as pd

from indsl.resample.auto_align import auto_align
from indsl.type_check import check_types

from ..exceptions import UserValueError


@check_types
def add(a, b, align_timesteps: bool = False):
    """Add.

    Add any two time series or numbers.

    Args:
        a: Time-series or number.
        b: Time-series or number.
        align_timesteps (bool) : Auto-align
            Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Time series.
    """
    a, b = auto_align([a, b], align_timesteps)
    return op.add(a, b)


@check_types
def sub(a, b, align_timesteps: bool = False):
    """Subtraction.

    The difference between two time series or numbers.

    Args:
        a: Time-series or number.
        b: Time-series or number.
        align_timesteps (bool) : Auto-align
           Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Time series.
    """
    a, b = auto_align([a, b], align_timesteps)
    return op.sub(a, b)


@check_types
def mul(a, b, align_timesteps: bool = False):
    """Multiplication.

    Multiply two time series or numbers.

    Args:
        a: Time-series or number.
        b: Time-series or number.
        align_timesteps (bool): Auto-align
           Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Time series.
    """
    a, b = auto_align([a, b], align_timesteps)
    return op.mul(a, b)


@check_types
def div(a, b, align_timesteps: bool = False):
    """Division.

    Divide two time series or numbers. If the time series in the
    denominator contains zeros, all instances are dropped from the final
    result.

    Args:
        a: Numerator
        b: Denominator
        align_timesteps (bool): Auto-align.
           Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Time series.
    """
    a, b = auto_align([a, b], align_timesteps)

    if type(b) is pd.Series:
        res = op.truediv(a, b).replace([np.inf, -np.inf], np.nan).dropna()
    elif type(b) is np.array:
        b = b.astype("float")  # Make sure it is a float to replace zeros (int) by np.nan (float)
        b[b == 0] = np.nan
        res = op.truediv(a, b)
        res = res[~np.isnan(res)]
    else:
        res = op.truediv(a, b)

    return res


@check_types
def power(a, b, align_timesteps: bool = False):
    """Power.

    Power of time series or numbers.

    Args:
        a: base time series or number
        b: exponent time series or number
        align_timesteps (bool): Auto-align
           Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Time series.
    """
    a, b = auto_align([a, b], align_timesteps)
    return op.pow(a, b)


def inv(x):
    """Inverse.

    Element-wise inverse of time series or numbers.

    Args:
        x: time series or numbers

    Returns:
        pandas.Series: Time series.
    """
    return 1 / x


def sqrt(x):
    """Square root.

    Square root of time series or numbers.

    Args:
        x: time series or numbers

    Returns:
        pandas.Series: Time series.
    """
    return np.sqrt(x)


def neg(x):
    """Negation.

    Negation of time series or numbers.

    Args:
        x: time series or numbers

    Returns:
        pandas.Series: Time series.
    """
    return op.neg(x)


def absolute(x):
    """Absolute value.

    The absolute value of time series or numbers.

    Args:
        x: time series or numbers

    Returns:
        pandas.Series: Time series.
    """
    return op.abs(x)


@check_types
def mod(a, b, align_timesteps: bool = False):
    """Modulo.

    Modulo of time series or numbers.

    Args:
        a: dividend time series or number
        b: divisor time series or number
        align_timesteps (bool): Auto-align
           Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Time series.
    """
    a, b = auto_align([a, b], align_timesteps)
    return op.mod(a, b)


@check_types
def arithmetic_mean(a, b, align_timesteps: bool = False):
    """Arithmetic mean.

    The mean of two time series or numbers.

    Args:
        a: Time series or number
        b: Time series or number
        align_timesteps (bool): Auto-align
           Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Time series.
    """
    return op.truediv(add(a, b, align_timesteps), 2)


@check_types
def arithmetic_mean_many(data: List[Union[pd.Series, float]], align_timesteps: bool = False) -> Union[pd.Series, float]:
    """Arithmetic mean many.

    The mean of multiple time series.

    Args:
        data: List of time series
        align_timesteps (bool): Auto-align
           Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Time series.
    """
    n = len(data)
    if n == 0:
        raise UserValueError("Expected at least one item in data parameter, got zero instead.")
    timeseries_sum = 0
    for new_ts in data:
        timeseries_sum = add(timeseries_sum, new_ts, align_timesteps)
    return op.truediv(timeseries_sum, n)



# Copyright 2021 Cognite AS
from __future__ import annotations

from typing import Literal, Optional, Union

import pandas as pd

from indsl.data_quality.low_density_identification import (
    low_density_identification_iqr,
    low_density_identification_modified_z_scores,
    low_density_identification_threshold,
    low_density_identification_z_scores,
)
from indsl.data_quality.score.base import DataQualityScore, DataQualityScoreAnalyser

# from indsl.exceptions import UserValueError
from indsl.type_check import check_types


class DensityDataQualityScoreAnalyser(DataQualityScoreAnalyser):
    """Density based data quality scores."""

    @check_types
    def __init__(self, series: pd.Series):
        """Density data quality score analyser init function.

        Args:
            series: Series to be analysed

        Raises:
            UserValueError: If the series has less than 2 values
            UserValueError: If series has no time index
        """
        super().__init__(series)
        self._low_density_detection_methods = {
            "iqr": low_density_identification_iqr,
            "z_scores": low_density_identification_z_scores,
            "modified_z_scores": low_density_identification_modified_z_scores,
            "threshold": low_density_identification_threshold,
        }

    @check_types
    def _get_low_density(
        self,
        analysis_start: pd.Timestamp,
        analysis_end: pd.Timestamp,
        low_density_detection_method: Literal["iqr", "z_scores", "modified_z_scores", "threshold"] = "iqr",
        **low_density_detection_options: Optional[Union[pd.Timedelta, int, bool]],
    ):
        method = self._low_density_detection_methods[low_density_detection_method]
        low_density_results = method(self.series, **low_density_detection_options)

        events = self._convert_series_to_events(low_density_results)
        events = self._filter_events_outside_analysis_period(events, analysis_start, analysis_end)
        # The first and last detected low density periods might range outside the analysis period. Let's fix this...
        events = self._limit_first_and_last_events_to_analysis_period(events, analysis_start, analysis_end)

        return events

    @check_types
    def compute_score(
        self,
        analysis_start: pd.Timestamp,
        analysis_end: pd.Timestamp,
        low_density_detection_method: Literal["iqr", "z_scores", "modified_z_scores", "threshold"] = "iqr",
        **low_density_detection_options: Optional[Union[pd.Timedelta, int, bool]],
    ) -> DataQualityScore:
        """Compute the low density analysis score.

        Args:
            analysis_start: Analyis start time
            analysis_end: Analyis end time
            low_density_detection_method: Low density detection method
                Must be one of "iqr", "z_scores", "modified_z_scores", "threshold". Default to "iqr".
            low_density_detection_options: Arguments to low density detection method
                Provided as a keyword dictionary

        Returns:
            DataQualityScore: A DataQualityScore object

        Raises:
            UserValueError: If analysis_start < analysis_end
            UserValueError: If the analysis start and end timestamps are outside the range of the series index
        """
        # initialize
        super().compute_score(analysis_start, analysis_end)
        # Treat empty series as one low density period
        if len(self.series) == 0:
            return DataQualityScore(analysis_start, analysis_end, [(analysis_start, analysis_end)])

        events = self._get_low_density(
            analysis_start, analysis_end, low_density_detection_method, **low_density_detection_options
        )

        return DataQualityScore(analysis_start, analysis_end, [tuple(event) for event in events])



# Copyright 2022 Cognite AS
import operator as op
import warnings

from enum import Enum
from typing import Literal, Optional

import pandas as pd

from scipy import signal

from indsl import versioning
from indsl.exceptions import UserTypeError, UserValueError
from indsl.ts_utils.ts_utils import fill_gaps, get_fixed_freq, is_na_all
from indsl.type_check import check_types
from indsl.validations import validate_series_has_time_index


@versioning.register(version="1.0", deprecated=True)
@check_types
def resample(
    data: pd.Series,
    method: Literal["fourier", "polyphase", "interpolate", "min", "max", "sum", "count", "mean"] = "fourier",
    granularity_current: Optional[str] = None,
    granularity_next: str = "1s",
    num: Optional[int] = None,
    downsampling_factor: Optional[int] = None,
    interpolate_resolution: Optional[str] = None,
    ffill_resolution: Optional[str] = None,
) -> pd.Series:
    """Resample.

    This method offers a robust filling of missing data points and data resampling of a given sampling frequency. Multiple
    data resampling options are available:

        * Fourier
        * Polynomial phase filtering
        * Linear interpolation (for up-sampling)
        * Min, max, sum, count, mean (for down-sampling)

    Args:
        data: Time series.

        method: Method.
            Resampling method:

                * "fourier" for Fourier method (default)
                * "polyphase" for polyphase filtering
                * "interpolate" for linear interpolation when upsampling
                * "min", "max", "sum", "count", "mean" when downsampling

        granularity_current: Current temporal resolution.
            Temporal resolution of uniform time series before resampling. Defaults to None.
            If not specified, the frequency will be implied, which only works if no data is missing.
            Follows Pandas DateTime convention.

        granularity_next: Final temporal resolution.
            Temporal resolution of uniform time series after resampling. Defaults to "1s".
            Either "Number of samples" or "Final temporal resolution" must be provided.

        num: Number of Samples.
            The number of samples in the resampled signal. If this is set, the time deltas will be inferred. Defaults
            to None. Either "Number of Samples" or "Final temporal resolution" must be provided.

        downsampling_factor: Down-sampling factor.
            The down-sampling factor is required for the polyphase filtering. Defaults to None.

        interpolate_resolution: Interpolation threshold.
            Gaps smaller than threshold will be interpolated, larger than this will be filled by noise.
            Defaults to None.

        ffill_resolution: Forward fill threshold.
            Gaps smaller than this threshold will be forward filled. Defaults to None.

    Returns:
        pandas.Series: Interpolated time series
        Uniform, resampled time series with specified number of data points.

    Raises:
        UserTypeError: data is not a time series
        UserTypeError: Either num or granularity_next has to be set
        UserTypeError: If specified, outside_fill must be either 'None' or 'extrapolate'.
        UserTypeError: Method has to be in 'fourier', 'polyphase', 'interpolate', 'min', 'max', 'sum', 'count'
        UserTypeError: Empty data time series
        UserTypeError: All values in the time series are NaN
        Warning: Can't infer time series resolution with missing data. Please provide resolution
    """
    if not granularity_next and (not num or num <= 0):
        raise UserTypeError("Either num or granularity_next has to be set.")

    validate_series_has_time_index(data)
    if len(data) == 0:
        raise UserValueError(f"Expected data to be of length > 0, got length {len(data)}")
    elif is_na_all(data):
        raise UserTypeError("All values in the time series are NaN.")

    # To resample data to uniform distribution
    if not granularity_current:
        # it returns none if it isn't able to infer the resolution
        granularity_current = pd.infer_freq(data.index, True)

        if not granularity_current:
            # TODO: pick max resolution and apply to the rest of the timeseries?
            warnings.warn(
                "Can't infer time series resolution with missing data. Please provide resolution", UserWarning
            )
            return data

    # make sure that it is uniform
    data = data.asfreq(freq=granularity_current)

    # add 1 if number is missing, in order to help Timedelta read it
    granularity_current = (
        granularity_current if any(char.isdigit() for char in granularity_current) else "1" + granularity_current
    )

    # remove nan
    data = fill_gaps(
        data=data,
        granularity=granularity_current,
        ffill_resolution=ffill_resolution,
        interpolate_resolution=interpolate_resolution,
    )

    if not num:
        num = 1 + int((data.index.max() - data.index.min()) / pd.Timedelta(granularity_next))

    if method == "fourier":
        series = pd.Series(signal.resample(data, num), index=pd.date_range(data.index.min(), data.index.max(), num))
    elif method == "polyphase":
        if not downsampling_factor:
            downsampling_factor = len(data)
        sig = signal.resample_poly(data, num, downsampling_factor)
        series = pd.Series(sig, index=pd.date_range(data.index.min(), data.index.max(), len(sig)))
    elif method == "interpolate":
        series = data.resample(granularity_next).interpolate()
    else:
        series = data.resample(granularity_next, origin="epoch")
        series = op.methodcaller(method)(series)

    return series


class AggregateEnum(Enum):  # noqa
    MEAN = "mean"
    INTERPOLATION = "interpolation"
    STEP_INTERPOLATION = "stepInterpolation"
    MAX = "max"
    MIN = "min"
    COUNT = "count"
    SUM = "sum"


@versioning.register(version="1.0", deprecated=True)
@check_types
def resample_to_granularity(
    series: pd.Series, granularity: str = "1h", aggregate: AggregateEnum = AggregateEnum.MEAN
) -> pd.Series:
    """Resample to granularity.

    Resample time series to a given fixed granularity (time delta) and aggregation type
    (`read more about aggregation <https://docs.cognite.com/dev/concepts/aggregation/>`_)

    Args:
        series: Time series.

        granularity: Granularity.
            Granularity defines the time range that each aggregate is calculated from. It consists of a time unit and a
            size. Valid time units are day or d, hour h, minute or m, and second or s. For example, 2h means that each
            time range should be 2 hours wide, 3m means 3 minutes, and so on.

        aggregate: Aggregate.
            Type of aggregation to use when resampling. Possible options are:

                * mean
                * max
                * min
                * count
                * sum

    Returns:
        pandas.Series: Resampled time series.
    """
    validate_series_has_time_index(series)
    if len(series) == 0:
        raise UserValueError(f"Expected series to be of length > 0, got length {len(series)}")

    # Translates from cdf aggregates to pandas method names:
    # TODO: 'mean' (point-weighted) does not equal CDF average (time-weighted).
    # Read more in the docs: https://docs.cognite.com/dev/concepts/aggregation/
    start_time = series.index[0]
    end_time = series.index[-1]
    if aggregate in (AggregateEnum.MIN, AggregateEnum.MAX, AggregateEnum.SUM, AggregateEnum.COUNT, AggregateEnum.MEAN):
        series = series.resample(granularity.replace("m", "T"), origin="epoch")
        series = op.methodcaller(aggregate.value)(series)
        if aggregate is AggregateEnum.MEAN:
            # We only fill inner holes with lin.interp for average:
            series = series.interpolate("slinear")
        return series.dropna()

    elif aggregate in (AggregateEnum.INTERPOLATION, AggregateEnum.STEP_INTERPOLATION):
        # We adhere to the CDF definition of interp. and step interp:
        fixed_freq = get_fixed_freq(start_time, end_time, granularity)
        series = series.reindex(series.index.union(fixed_freq))  # Note: reindex, not resample
        if aggregate is AggregateEnum.INTERPOLATION:
            return series.interpolate("slinear")[fixed_freq].dropna()
        else:
            return series.ffill().bfill()[fixed_freq]



# Copyright 2022 Cognite AS
from typing import Literal

import numpy as np
import pandas as pd

from sklearn.linear_model import Lasso, LinearRegression, Ridge
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures

from ..exceptions import UserValueError
from ..type_check import check_types


@check_types
def poly_regression(
    data: pd.Series,
    order: int = 2,
    method: Literal["Lasso", "Ridge", "No regularisation"] = "No regularisation",
    alpha: float = 0.1,
) -> pd.Series:
    """Polynomial.

    Fit a polynomial curve of a specified degree to the data. Default method corresponds to a ordinary least squares
    fitting procedure but method can be changed to allow for L1 or L2 regularisation.

    Args:
        data: Time series.
            Data to fit the polynomial regression
        order: Polynomial order.
        method: Method.
            Type of regularisation to apply (Lasso or Ridge). Default is simple linear least squares with no regularisation.
        alpha: Alpha.
            Only applies to either Ridge or Lasso methods which sets the penalty for either L2 or L1 regularisation.
            Value of 0 means that there is no penalty and this essentially equivalent to ordinary least squares.

    Returns:
        pd.Series: Fitted data.
    """
    # Check alpha
    if not 0 < alpha < 1:
        raise UserValueError("Alpha needs to be a float between 0 and 1")

    # Select method and build pipeline
    if method == "No regularisation":
        fitter = LinearRegression()
    elif method == "Ridge":
        fitter = Ridge(alpha=alpha)
    elif method == "Lasso":
        fitter = Lasso(alpha=alpha)

    model = make_pipeline(PolynomialFeatures(order), fitter)

    # Transform variables
    data = data.replace([np.inf, -np.inf], np.nan).dropna()

    if len(data) < (order + 1):
        raise UserValueError(
            f"Not enough data (got {len(data)} values) to perform operation (min {order + 1} values required!)"
        )

    x = (np.array(data.index, dtype=np.int64) - data.index[0].value) / 1e9
    y = data.to_numpy()

    # Fit, transform and return result
    model.fit(x.reshape(-1, 1), y)
    y_pred = model.predict(x.reshape(-1, 1))
    return pd.Series(y_pred, index=data.index)
